import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;

public class ANB_BDeu {
	public int fANB; //learning with ANB constraint if fABN = 1, learning GBN otherwise.
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public int CrossValidation; //How many times split data for cross validation.
	public ArrayList<int[]> RD; //Read data
	public ArrayList<int[]> LD; //Learning data for each validation.
	public getLocalScore gls;
	public getBestParents gbp;
	public getBestNets gbn;
	public ArrayList<ArrayList<Integer>> structure;

	public static void main(String[] args) throws IOException, InterruptedException {
		String a = "data/mux6.csv";
		String b = "data/a.txt";
		String c = "data";
		new ANB_BDeu(a, c, 10);
	}
	
	public ANB_BDeu(String path_to_dataset, String path_to_directory_for_output, int cv) throws IOException{
		System.out.print("Exeute ANB-BDeu with ");
		System.out.println(cv + "-fold cross-validation.");
		CrossValidation = cv;
		fANB = 1;
		double alpha = 1.0;// equivalent sample size
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);
			gls = new getLocalScore(fANB, LD, NV, NC, MNC, alpha);
			gbp = new getBestParents(fANB, NV, NC, MNC, gls.LS, false);
			gbn = new getBestNets(fANB, NV, NC, MNC, gls.LS, gbp.BP);
			structure = gbn.learnedStructure;
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	protected void readData(String DataBaseName) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(DataBaseName)));
		ArrayList<String[]> tRD = new ArrayList<String[]>();
		String line = br.readLine();
		NV = line.split(",").length;
		while(line != null){
			String[] data = line.split(",").clone();
			tRD.add(data);
			line = br.readLine();
		}
		int[] max = new int[NV];
		RD = new ArrayList<int[]>();
		for(int i = 0; i < tRD.size(); ++i){
			RD.add(new int[NV]);
			for(int j = 0; j < NV; ++j){
				RD.get(i)[j] = Integer.parseInt(tRD.get(i)[j]);
				if(max[j] < RD.get(i)[j]){
					max[j] = RD.get(i)[j];
				}
			}
		}
		MNC = 0;
		NC = new int[NV];
		for(int i = 0; i < NV; ++i){
			NC[i] = max[i] + 1;
			if(MNC < NC[i]){
				MNC = NC[i];
			}
		}
		br.close();
	}

	protected void readLearningData(int j) throws IOException{
		if(CrossValidation == 1){
			LD = RD;
			return;
		}
		LD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = 0; i < RD.size(); ++i){
				if(i < j*(RD.size()/CrossValidation) || (j+1)*(RD.size()/CrossValidation) <= i)
					LD.add(RD.get(i));
			}
		}
		else{
			for(int i = 0; i < (CrossValidation - 1)*(RD.size()/CrossValidation); ++i){
				LD.add(RD.get(i));
			}
		}
	}

	public void writeStructure(String path, int j){
		try{
			File outputDir = new File(path);
			if (outputDir.exists()){
				File file = new File(path + "/" + (j+1) + ".txt");
				file.createNewFile();
				PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
				for(int i = 0; i < NV; ++i){
					pw.print(i);
					pw.print("\t");
				}
				pw.println();
				for(int i = 0; i < NV; ++i){
					ArrayList<Integer> t = structure.get(i);
					if(t.size() != 0){
						pw.print(t.get(0));
						for(int p = 1; p < t.size(); ++p){
							pw.print(",");
							pw.print(t.get(p));
						}
					}else{
						pw.print(-1);
					}
					pw.print("\t");
				}
				pw.close();
			}else{
				System.out.println();
				System.out.println("The output directory not found.");
				System.exit(0);
			}
		}catch(IOException e){
			System.out.println(e);
		}
	}
}
