import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Queue;
import java.util.Random;
import java.util.Set;

public class BNC2P {
	public static int NV;
	public static int[] NC;
	public static int MNC;
	public static ArrayList<int[]> LD;

	public static int COUNT;
	public static int CrossValidation;
	public ArrayList<int[]> RD;
	public int fANB;
	public ArrayList<ArrayList<Integer>> structure;
	public double[][][] scoreList;
	public int end_condition = 10000000;

	public BNC2P(String path_to_dataset, String path_to_directory_for_output, int cv) throws IOException, InterruptedException{
		System.out.print("Exeute BNC2P with ");
		System.out.println(cv + "-fold cross-validation.");
		CrossValidation = cv;
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);
			ArrayList<ArrayList<ArrayList<Integer>>> candidateStrs = new ArrayList<ArrayList<ArrayList<Integer>>>();
			structure = new ArrayList<>();
			for(int i = 0; i < NV; ++i){
				structure.add(new ArrayList<Integer>());
			}
			double s;
			while(true){
				s = LearningGreedyCLL_parallel_P2(structure);
				if(s == end_condition) break;
			}
			candidateStrs.add(structure);

			for(int m = 0; m < 10; ++m){
				structure = makingStructureP2();
				while(true){
					s = LearningGreedyCLL_parallel_P2(structure);
					if(s == end_condition) break;
				}
				candidateStrs.add(structure);
			}
			double max = -Math.pow(10, 10);
			int maxStr = 0;
			for(int m = 0; m < candidateStrs.size(); ++m){
				structure = candidateStrs.get(m);
				ArrayList<HashMap<BigInteger, int[]>> tn = makeCft(structure);
				double score = calcCLL(structure, tn);
				if(max < score){
					max = score;
					maxStr = m;
				}
			}
			structure = candidateStrs.get(maxStr);
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	public class GD implements Runnable {
		private double[][][] scorelist;
		private ArrayList<ArrayList<Integer>> str;
		private int p1;
		private int p2;
		private int type;
		public GD(double[][][] scorelist, ArrayList<ArrayList<Integer>> str, int type, int p1, int p2) { // p1->p2
			this.scorelist = scorelist;
			this.str = str;
			this.p1 = p1;
			this.p2 = p2;
			this.type = type;
		}
		@Override
		public void run() {
			ArrayList<ArrayList<Integer>> tstr = makeNeighborhood(str, type, p1, p2);
			ArrayList<HashMap<BigInteger, int[]>> tn = makeCft(tstr);
			scorelist[type][p2][p1] = calcCLL(tstr, tn);
		}
	}

	private ArrayList<ArrayList<Integer>> makeNeighborhood(ArrayList<ArrayList<Integer>> str, int type, int p1, int p2){
		ArrayList<ArrayList<Integer>> tstr = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			tstr.add((ArrayList<Integer>) str.get(i).clone());
		}
		switch(type){
		case 0:
			tstr.get(p2).add(p1);
			Collections.sort(tstr.get(p2));
			break;
		case 1:
			tstr.get(p2).remove(tstr.get(p2).indexOf(p1));
			break;
		case 2:
			tstr.get(p1).add(p2);
			Collections.sort(tstr.get(p1));
			tstr.get(p2).remove(tstr.get(p2).indexOf(p1));
			break;
		}
		return tstr;
	}

	void scoreListInit(){
		scoreList = new double[3][NV][NV];
		for(int i = 0; i < 3; ++i){
			for(int j = 0; j < NV; ++j){
				for(int k = 0; k < NV; ++k){
					scoreList[i][j][k] = -1000000000;
				}
			}
		}
	}

	protected void readData(String DataBaseName) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(DataBaseName)));
		ArrayList<String[]> tRD = new ArrayList<String[]>();
		String line = br.readLine();
		NV = line.split(",").length;
		while(line != null){
			String[] data = line.split(",").clone();
			tRD.add(data);
			line = br.readLine();
		}

		int[] max = new int[NV];
		RD = new ArrayList<int[]>();
		for(int i = 0; i < tRD.size(); ++i){
			RD.add(new int[NV]);
			for(int j = 0; j < NV; ++j){
				RD.get(i)[j] = Integer.parseInt(tRD.get(i)[j]);
				if(max[j] < RD.get(i)[j]){
					max[j] = RD.get(i)[j];
				}
			}
		}
		MNC = 0;
		NC = new int[NV];
		for(int i = 0; i < NV; ++i){
			NC[i] = max[i] + 1;
			if(MNC < NC[i]){
				MNC = NC[i];
			}
		}
		br.close();
	}

	protected void readLearningData(int j) throws IOException{
		if(CrossValidation == 1){
			LD = RD;
			return;
		}
		LD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = 0; i < RD.size(); ++i){
				if(i < j*(RD.size()/CrossValidation) || (j+1)*(RD.size()/CrossValidation) <= i)
					LD.add(RD.get(i));
			}
		}
		else{
			for(int i = 0; i < (CrossValidation - 1)*(RD.size()/CrossValidation); ++i){
				LD.add(RD.get(i));
			}
		}
	}

	public void writeStructure(String path, int j){
		try{
			File outputDir = new File(path);
			if (outputDir.exists()){
				File file = new File(path + "/" + (j+1) + ".txt");
				file.createNewFile();
				PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
				for(int i = 0; i < NV; ++i){
					pw.print(i);
					pw.print("\t");
				}
				pw.println();
				for(int i = 0; i < NV; ++i){
					ArrayList<Integer> t = structure.get(i);
					if(t.size() != 0){
						pw.print(t.get(0));
						for(int p = 1; p < t.size(); ++p){
							pw.print(",");
							pw.print(t.get(p));
						}
					}else{
						pw.print(-1);
					}
					pw.print("\t");
				}
				pw.close();
			}else{
				System.out.println();
				System.out.println("The output directory not found.");
				System.exit(0);
			}
		}catch(IOException e){
			System.out.println(e);
		}
	}

	private double LearningGreedyCLL_parallel_P2(ArrayList<ArrayList<Integer>> parents) throws InterruptedException{ // greedy search
		scoreListInit();
		ArrayList<Thread> thrList = new ArrayList<>();
		for(int i = 0; i < NV - fANB; ++i){
			//add
			for(int j = 0; j < NV - fANB; ++j){
				if(i == j) continue;
				ArrayList<Integer> parent = parents.get(i);
				if((parents.get(i).size() < 2) && !parent.contains(j) && !judgeLoop(i, j, parents)){
					Runnable runnable = new GD(scoreList, parents, 0, j, i);
					Thread thr = new Thread(runnable);
					thr.start();
					thrList.add(thr);
				}
			}
			//delete
			for(int j = 0; j < NV - fANB; ++j){
				if(i == j) continue;
				ArrayList<Integer> parent = parents.get(i);
				if(parent.contains(j)){
					Runnable runnable = new GD(scoreList, parents, 1, j, i);
					Thread thr = new Thread(runnable);
					thr.start();
					thrList.add(thr);
				}
			}
			//reverse
			for(int j = 0; j < NV - fANB; ++j){
				if(i == j) continue;
				ArrayList<Integer> parent = parents.get(i);
				if((parents.get(j).size() < 2) && parent.contains(j) && !judgeLoop_reverse(i, j, parents)){
					Runnable runnable = new GD(scoreList, parents, 2, j, i);
					Thread thr = new Thread(runnable);
					thr.start();
					thrList.add(thr);
				}
			}
		}

		for(int c = 0; c < thrList.size(); ++c){
			thrList.get(c).join();
		}

		ArrayList<HashMap<BigInteger, int[]>> n1 = makeCft(parents);
		double max = calcCLL(parents, n1);

		int type = -1;
		int p1 = 0;
		int p2 = 0;
		for(int i = 0; i < 3; ++i){
			for(int j = 0; j < NV; ++j){
				for(int k = 0; k < NV; ++k){
					if(max < scoreList[i][j][k]){
						max = scoreList[i][j][k];
						type = i;
						p2 = j;
						p1 = k;
					}
				}
			}
		}

		ArrayList<Integer> parent = parents.get(p2);
		switch(type){
		case -1: return 10000000;
		case 0: parent.add(p1); Collections.sort(parent); break;
		case 1: parent.remove(parent.indexOf(p1)); break;
		case 2:
			parent.remove(parent.indexOf(p1));
			parents.get(p1).add(p2);
			Collections.sort(parents.get(p1));
			break;
		}
		return max;
	}

	private boolean judgeLoop(int in, int out, ArrayList<ArrayList<Integer>> parents){ // in <- out
		Set<Integer> mile = new HashSet();
		Queue<Integer> candidate = new ArrayDeque();
		candidate.add(out);
		while(!candidate.isEmpty()){
			int now = candidate.poll();
			if(now == in){
				return true;
			}
			for(int i = 0; i < parents.get(now).size(); ++i){
				if(!mile.contains(parents.get(now).get(i))){
					candidate.add(parents.get(now).get(i));
				}
			}
			mile.add(now);
		}
		return false;
	}

	private boolean judgeLoop_reverse(int in, int out, ArrayList<ArrayList<Integer>> parents){// reverse of (in <- out)
		Set<Integer> mile = new HashSet();
		Queue<Integer> candidate = new ArrayDeque();
		for(int i = 0; i < parents.get(in).size(); ++i){
			if(parents.get(in).get(i) != out){
				candidate.add(parents.get(in).get(i));
			}
		}
		mile.add(in);
		while(!candidate.isEmpty()){
			int now = candidate.poll();
			if(now == out){
				return true;
			}
			for(int i = 0; i < parents.get(now).size(); ++i){
				if(!mile.contains(parents.get(now).get(i))){
					candidate.add(parents.get(now).get(i));
				}
			}
			mile.add(now);
		}
		return false;
	}

	private double calcCLL(ArrayList<ArrayList<Integer>> parents, ArrayList<HashMap<BigInteger, int[]>> n){
		double L = 0.0;
		for(int d = 0; d < LD.size(); ++d){
			int save = LD.get(d)[NV - 1];
			double[][] p = new double[NV][NC[NV - 1]];
			int[] data = new int[NV];
			for(int i = 0; i < NV; ++i){
				data[i] = LD.get(d)[i];
			}
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				for(int i = 0; i < NV; ++i){
					BigInteger key = makeKey(data, i, parents);
					if(n.get(i).containsKey(key)){
						double n_ij = 0;
						for(int k = 0; k < NC[i]; ++k){
							n_ij += n.get(i).get(key)[k];
						}
						double n_ijk = n.get(i).get(key)[data[i]];
						p[i][c] = n_ijk / n_ij;
					}else{
						p[i][c] = 0;
					}
				}
			}
			double total = 0.0;
			double[] lc = new double[NC[NV - 1]];
			for(int v = 0; v < NC[NV - 1]; ++v){
				boolean pzero = false;
				lc[v] = 0.0;
				for(int i = 0; i < NV; ++i){
					if(p[i][v] != 0){
						lc[v] += Math.log(p[i][v]);
					}else{
						pzero = true;
						break;
					}
				}
				if(!pzero){
					total += Math.exp(lc[v]);
				}
			}
			L += lc[save] - Math.log(total);
		}

		return L;
	}

	private ArrayList<HashMap<BigInteger, int[]>> makeCft(ArrayList<ArrayList<Integer>> parents){
		ArrayList<HashMap<BigInteger, int[]>> n = new ArrayList<HashMap<BigInteger, int[]>>();
		for(int i = 0; i < NV; ++i) n.add(new HashMap<BigInteger, int[]>());
		for(int l=0; l<LD.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int i = 0; i < NV; ++i) currentLearnData[i] = LD.get(l)[i];
			for(int i = 0; i < NV; i++){
				BigInteger hash = makeKey(currentLearnData, i, parents);
				if(n.get(i).containsKey(hash)){
					(n.get(i).get(hash)[currentLearnData[i]])++;
				}else{
					int[] t = new int[NC[i]];
					for(int j = 0; j < NC[i]; ++j){
						t[j] = 0;
					}
					t[currentLearnData[i]] = 1;
					n.get(i).put(hash, t);
				}
			}
		}
		return n;
	}

	private ArrayList<ArrayList<Integer>> makingStructureP2(){
		ArrayList<ArrayList<Integer>> str = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			str.add(new ArrayList<Integer>());
		}
		ArrayList<Integer> order = randomOrder(NV);

		Random rnd = new Random();
		for(int i = 0; i < NV; ++i){
			ArrayList<Integer> t = randomOrder(i);
			int pnum = (i >= 3) ? rnd.nextInt(3) : rnd.nextInt(i+1);
			for(int j = 0; j < pnum; ++j){
				str.get(order.get(i)).add(order.get(t.get(j)));
			}

			Collections.sort(str.get(order.get(i)));
		}
		return str;
	}

	private ArrayList<Integer> randomOrder(int size){
		ArrayList<Integer> r = new ArrayList<>();
		for(int i = 0; i < size; ++i){
			r.add(i);
		}
		Collections.shuffle(r);
		return r;
	}

	private BigInteger makeKey(int[] data, int i, ArrayList<ArrayList<Integer>> parentList){
		if(!parentList.get(i).isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parentList.get(i).size(); ++p){
				int t = parentList.get(i).get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	private BigInteger product(int a, int b){
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}
}
