import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;

public class EvaluateClassificationAccuracy {
	public int fANB; //learning ANB if fABN = 1, learning GBN otherwise.
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public int CrossValidation; //How many times split data for cross validation.
	public ArrayList<int[]> RD; //Read data
	public ArrayList<int[]> LD; //Learning data for each validation.
	public ArrayList<int[]> TD; //Test data for each validation.
	public String[] SD;
	public getLocalScore gls;
	public getBestParents gbp;
	public getBestNets gbn;
	public ArrayList<ArrayList<Integer>> structure;
	public ArrayList<double[]> prob;

	public static void main(String[] args) throws IOException, InterruptedException {
		int cv = Integer.valueOf(args[3]);
		new EvaluateClassificationAccuracy(args[0], args[1], args[2], cv);
	}

	public EvaluateClassificationAccuracy(String path_to_dataset, String path_to_directory_for_structures, String path_to_output,int cv) throws IOException{
		System.out.print("Evaluate the average classification accuracy from ");
		System.out.println(cv + "-fold cross-validation.");
		CrossValidation = cv;
		readData(path_to_dataset);
		double count = 0;
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);
			readTestData(j);
			initProb();
			readStructureData(path_to_directory_for_structures + "/"+ (j+1) + ".txt");
			Classification(LD, TD, 1.0, structure);
			count += calcCorrect(TD);
			System.out.println("; Complete.");
		}
		double aveAccuracy = count/(double)RD.size();
		System.out.println("Average accuracy: " + aveAccuracy);
		writeAccuracy(path_to_output, aveAccuracy);
	}

	private void initProb(){
		prob = new ArrayList<double[]>();
		for(int t = 0; t < TD.size(); ++t){
			double[] c = new double[NC[NV - 1]];
			for(int v = 0; v < NC[NV - 1]; ++v){
				c[v] = 0.0;
			}
			prob.add(c);
		}
	}

	private void readStructureData(String path) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(path)));
		String line = br.readLine();
		line = br.readLine();
		SD = line.split("\t").clone();
		br.close();
		structure = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			structure.add(new ArrayList<Integer>());
			if(!SD[i].equals("-1")){
				String[] parent = SD[i].split(",");
				int pa_length = parent.length;
				for(int pa=0; pa< pa_length ; ++pa){
					structure.get(i).add(Integer.parseInt(parent[pa]));
				}
			}
		}
	}

	public void Classification(ArrayList<int[]> ld, ArrayList<int[]> td, double alpha, ArrayList<ArrayList<Integer>> parentList){
		//init Prob
		prob = new ArrayList<double[]>();
		for(int t = 0; t < td.size(); ++t){
			double[] c = new double[NC[NV - 1]];
			for(int v = 0; v < NC[NV - 1]; ++v){
				c[v] = 0.0;
			}
			prob.add(c);
		}

		ArrayList<HashMap<BigInteger, int[]>> n = makeCft(parentList, ld);
		for(int t = 0; t < td.size(); ++t){
			int[] currentTestData = new int[NV];
			for(int i = 0; i < NV; ++i) currentTestData[i] = td.get(t)[i];
			double[] p = new double[NC[NV - 1]];
			for(int i = 0; i < NC[NV - 1]; ++i) p[i] = 0.0;

			int i = 0;
			for(i = 0; i < NV; ++i){
				double alpha_ij = alpha;
				double denominator = 1.0;
				for(int j = 0; j < parentList.get(i).size(); ++j){
					denominator *= NC[parentList.get(i).get(j)];
				}
				alpha_ij /= denominator;
				double alpha_ijk = alpha / (denominator*NC[i]);

				for(int v = 0; v < NC[NV - 1]; ++v){
					currentTestData[NV - 1] = v;
					//System.out.println("makekey");
					BigInteger hash = makeKey(currentTestData, i, parentList);
					//System.out.println(hash);
					double n_ij = 0.0;
					for(int k = 0; k < NC[i]; ++k) n_ij += (n.get(i).containsKey(hash)) ? n.get(i).get(hash)[k] : 0;
					//if(v == 0) System.out.println(n_ij);
					double n_ijk = (n.get(i).containsKey(hash)) ? n.get(i).get(hash)[currentTestData[i]] : 0;
					p[v] += Math.log(n_ijk + alpha_ijk) - Math.log(n_ij + alpha_ij);
				}
			}
			double total = 0.0;
			for(int c = 0; c < NC[NV - 1]; ++c){
				total += Math.exp(p[c]);
				if(total == 0){
					total = 1;
				}
			}
			for(int j = 0; j < NC[NV - 1]; ++j){
				prob.get(t)[j] = Math.exp(p[j]) / total;
			}
		}
	}

	private double calcCorrect(ArrayList<int[]> td){
		double count = 0;
		for(int t = 0; t < td.size(); ++t){
			double tPred = 0.0;
			int testlabel = td.get(t)[NV - 1];
			boolean flag = false;
			for(int v = 0; v < testlabel; ++v){
				if(prob.get(t)[v] > prob.get(t)[testlabel]){
					flag = true;
				}else if(prob.get(t)[v] == prob.get(t)[testlabel]){
					++tPred;
				}
			}
			for(int v = testlabel + 1; v < NC[NV - 1]; ++v){
				if(prob.get(t)[v] > prob.get(t)[testlabel]){
					flag = true;
				}else if(prob.get(t)[v] == prob.get(t)[testlabel]){
					++tPred;
				}
			}
			if(!flag){
				if(tPred == 0){
					++count;
				}else{
					count += 1.0 / (tPred + 1.0);
				}
			}
		}
		return count;
	}

	private ArrayList<HashMap<BigInteger, int[]>> makeCft(ArrayList<ArrayList<Integer>> parents, ArrayList<int[]> ld){
		ArrayList<HashMap<BigInteger, int[]>> n = new ArrayList<HashMap<BigInteger, int[]>>();
		for(int i = 0; i < NV; ++i) n.add(new HashMap<BigInteger, int[]>());
		for(int l=0; l<ld.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int i = 0; i < NV; ++i) currentLearnData[i] = ld.get(l)[i];

			for(int i = 0; i < NV; i++){
				BigInteger hash = makeKey(currentLearnData, i, parents);
				if(n.get(i).containsKey(hash)){
					(n.get(i).get(hash)[currentLearnData[i]])++;
				}else{
					int[] t = new int[NC[i]];
					for(int j = 0; j < NC[i]; ++j){
						t[j] = 0;
					}
					t[currentLearnData[i]] = 1;
					n.get(i).put(hash, t);
				}
			}
		}
		return n;
	}

	private BigInteger makeKey(int[] data, int i, ArrayList<ArrayList<Integer>> parentList){
		if(!parentList.get(i).isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parentList.get(i).size(); ++p){
				int t = parentList.get(i).get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	private BigInteger product(int a, int b){
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}

	protected void readData(String DataBaseName) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(DataBaseName)));
		ArrayList<String[]> tRD = new ArrayList<String[]>();
		String line = br.readLine();
		NV = line.split(",").length;
		while(line != null){
			String[] data = line.split(",").clone();
			tRD.add(data);
			line = br.readLine();
		}
		int[] max = new int[NV];
		RD = new ArrayList<int[]>();
		for(int i = 0; i < tRD.size(); ++i){
			RD.add(new int[NV]);
			for(int j = 0; j < NV; ++j){
				RD.get(i)[j] = Integer.parseInt(tRD.get(i)[j]);
				if(max[j] < RD.get(i)[j]){
					max[j] = RD.get(i)[j];
				}
			}
		}
		MNC = 0;
		NC = new int[NV];
		for(int i = 0; i < NV; ++i){
			NC[i] = max[i] + 1;
			if(MNC < NC[i]){
				MNC = NC[i];
			}
		}
		br.close();
	}

	protected void readLearningData(int j) throws IOException{
		if(CrossValidation == 1){
			LD = RD;
			return;
		}
		LD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = 0; i < RD.size(); ++i){
				if(i < j*(RD.size()/CrossValidation) || (j+1)*(RD.size()/CrossValidation) <= i)
					LD.add(RD.get(i));
			}
		}
		else{
			for(int i = 0; i < (CrossValidation - 1)*(RD.size()/CrossValidation); ++i){
				LD.add(RD.get(i));
			}
		}
	}

	private void readTestData(int j) throws IOException{
		if(CrossValidation == 1){
			TD = RD;
			return;
		}
		TD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = j*(RD.size()/CrossValidation); i < (j+1)*(RD.size()/CrossValidation); ++i){
				TD.add(RD.get(i));
			}
		}
		else{
			for(int i = j*(RD.size()/CrossValidation); i < (j+1)*(RD.size()/CrossValidation) + RD.size()%CrossValidation; ++i){
				TD.add(RD.get(i));
			}
		}
	}

	public void writeAccuracy(String path, double accuracy){
		try{
			File file = new File(path);
			file.createNewFile();
			PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
			pw.println(accuracy);
			pw.close();
		}catch(IOException e){
			System.out.println(e);
		}
	}
}
