import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.math3.distribution.ChiSquaredDistribution;
import org.apache.commons.math3.special.Gamma;

public class FeatureSelection {
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public ArrayList<int[]> LD; //Learning data for each validation.
	public getLocalScore gls;
	public getBestParents gbp;
	public getBestNets gbn;
	public ArrayList<ArrayList<Integer>> structure;
	public ArrayList<Integer> fset;

	public FeatureSelection(int method, int nv, int[] nc, int mnc, ArrayList<int[]> ld){
		NV = nv;
		NC = nc;
		MNC = mnc;
		LD = ld;
		switch(method){
		case 2: fset = S2TMB(1.0); break;
		case 3: fset = MMPC(); break;
		case 4: fset = HITON_PC(); break;
		case 5: fset = PCMB(); break;
		default: System.out.println("Unknown method number"); System.exit(0); break;
		}
	}

	public FeatureSelection(double delta, double ess, int nv, int[] nc, int mnc, ArrayList<int[]> ld){
		NV = nv;
		NC = nc;
		MNC = mnc;
		LD = ld;
		fset = TheProposedPCsearch(delta, ess);
	}

	private ArrayList<Integer> TheProposedPCsearch(double thv, double ess){
		ArrayList<Integer> PC = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i){
			//make Cft for independent
			double[] cfti_ind = new double[NC[i]];
			//make Cft for dependent
			double[][] cfti_dep = new double[NC[NV - 1]][NC[i]];
			for(int d = 0; d < LD.size(); ++d){
				cfti_ind[LD.get(d)[i]]++;
				cfti_dep[LD.get(d)[NV - 1]][LD.get(d)[i]]++;
			}
			double ML_dep = 0;
			double alpha = ess;
			double alpha_ij = alpha / (double)NC[NV - 1];
			double alpha_ijk = alpha / (double)(NC[NV - 1] * NC[i]);
			for(int c = 0; c < NC[NV - 1]; ++c) {
				double n_ij = 0;
				for(int k = 0; k < NC[i]; ++k){
					double n_ijk = cfti_dep[c][k];
					ML_dep += Gamma.logGamma(alpha_ijk + n_ijk) - Gamma.logGamma(alpha_ijk);
					n_ij += n_ijk;
				}
				ML_dep += Gamma.logGamma(alpha_ij) - Gamma.logGamma(alpha_ij + n_ij);
			}
			double ML_ind = 0;
			alpha_ij = alpha;
			alpha_ijk = alpha / (double)(NC[i]);
			double n_ij = 0;
			for(int k = 0; k < NC[i]; ++k){
				double n_ijk = cfti_ind[k];
				ML_ind += Gamma.logGamma(alpha_ijk + n_ijk) - Gamma.logGamma(alpha_ijk);
				n_ij += n_ijk;
			}
			ML_ind += Gamma.logGamma(alpha_ij) - Gamma.logGamma(alpha_ij + n_ij);
			if(ML_ind > ML_dep + Math.log(thv)){

			}else{
				PC.add(i);
			}
		}
		return PC;
	}

	public ArrayList<Integer> S2TMB(double alpha){
		//PC set
		ArrayList<Integer> PCset = new ArrayList<>();
		ArrayList<Integer> candidate = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i) candidate.add(i);
		//make clone of LD
		ArrayList<int[]> saveLD = new ArrayList<>();
		for(int d = 0; d < LD.size(); ++d){
			saveLD.add(LD.get(d).clone());
		}
		while(!candidate.isEmpty()){
			int o = candidate.get(0);
			candidate.remove(0);
			ArrayList<Integer> Z = new ArrayList<Integer>();
			Z.add(o);
			Z.add(NV - 1);
			for(Integer pc: PCset){
				Z.add(pc);
			}
			Collections.sort(Z);
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[Z.size()];
				for(int i = 0; i < Z.size(); ++i){
					data[i] = LD.get(d)[Z.get(i)];
				}
				tLD.add(data);
			}
			int[] tNC = new int[Z.size()];
			for(int i = 0; i < Z.size(); ++i){
				tNC[i] = NC[Z.get(i)];
			}

			int saveNV = NV;
			int[] saveNC = NC.clone();
			NV = Z.size();
			LD = tLD;
			NC = tNC;


			gls = new getLocalScore(0, LD, NV, NC, MNC, alpha);
			gbp = new getBestParents(0, NV, NC, MNC, gls.LS, false);
			gbn = new getBestNets(0, NV, NC, MNC, gls.LS, gbp.BP);
			ArrayList<ArrayList<Integer>> tstructure = gbn.learnedStructure;
			structure = new ArrayList<>();
			for(int i = 0; i < saveNV; ++i) structure.add(new ArrayList<>());
			for(int i = 0; i < tstructure.size(); ++i){
				for(int pa = 0; pa < tstructure.get(i).size(); ++pa){
					structure.get(Z.get(i)).add(Z.get(tstructure.get(i).get(pa)));
				}
			}
			NV = saveNV;
			NC = saveNC;
			LD = saveLD;

			//update PC set
			PCset = new ArrayList<>();
			for(Integer pa: structure.get(NV - 1)){
				PCset.add(pa);
			}
			for(int i = 0; i < NV; ++i){
				if(structure.get(i).contains(NV - 1) && !PCset.contains(i)){
					PCset.add(i);
				}
			}
			gls = null;
			gbp = null;
			gbn = null;
		}
		return PCset;
	}

	private ArrayList<Integer> MMPC(){
		ArrayList<Integer> CPC = _MMPC();
		for(int i = 0; i < CPC.size(); ++i){
			//backup data
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[NV];
				for(int j = 0; j < NV; ++j){
					data[j] = LD.get(d)[j];
				}
				tLD.add(data);
			}
			int[] tNC = new int[NV];
			for(int j = 0; j < NV; ++j){
				tNC[j] = NC[j];
			}
			//learn attributes' PC
			arrangeClassVariable(CPC.get(i));
			ArrayList<Integer> attributesPC = _MMPC();
			if(!attributesPC.contains(NV - 2)){
				CPC.remove(i);
				--i;
			}
			LD = tLD;
			NC = tNC;
		}
		return CPC;
	}

	private ArrayList<Integer> _MMPC(){
		ArrayList<Integer> PC = new ArrayList<>();
		ArrayList<Integer> candidate = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i) candidate.add(i);
		// add true positive
		while(!candidate.isEmpty()){
			int variable_maxdep = 0;
			double maxdep = -Math.pow(10, 100);
			ArrayList<Integer> removedList = new ArrayList<>();
			for(int c = 0; c < candidate.size(); ++c){
				int i = candidate.get(c);
				if(PC.contains(i)) continue;
				double dep = getMindep(PC, i);
				if(-dep > 0.05){
					removedList.add(candidate.get(c));
				}
				if(dep > maxdep){
					maxdep = dep;
					variable_maxdep = i;
				}
			}
			for(int r: removedList){
				candidate.remove(candidate.indexOf(r));
			}
			if(-maxdep <= 0.05){
				PC.add(variable_maxdep);
				Collections.sort(PC);
				candidate.remove(candidate.indexOf(variable_maxdep));
			}else{
				break;
			}
		}

		//remove false positive

		for(int i = 0; i < PC.size(); ++i){
			boolean remove = remove_false_positive(PC, PC.get(i));
			if(remove) --i;
		}


		return PC;
	}

	private double getMindep(ArrayList<Integer> PC, int i){
		double min = Math.pow(10, 100);
		for(int j = 0; j < (1 << PC.size()); ++j){
			ArrayList<Integer> pattern = DecodeParents(j, PC);
			double num_allpattern = NC[i] * NC[NV - 1];
			for(int e: pattern){
				num_allpattern *= NC[e];
			}
			double dimension = (NC[i] - 1) * (NC[NV - 1] - 1);
			for(int e: pattern){
				dimension *= NC[e];
			}
			if(dimension == 0) dimension = 1;
			if(LD.size() / num_allpattern < 5){
				min = 1.0;
				break;
			}
			double statistics = G2_statistics(i, NV - 1, pattern);
			double pvalue = 1 - new ChiSquaredDistribution(dimension).cumulativeProbability(statistics);
			if(min > -pvalue) {
				min = -pvalue;
			}
		}

		return min;
	}

	private boolean remove_false_positive(ArrayList<Integer> PC, int i){
		for(int j = 0; j < (1 << (PC.size() - 1)); ++j){
			ArrayList<Integer> pattern = DecodeParents(i, j, PC);
			double num_allpattern = NC[i] * NC[NV - 1];
			for(int e: pattern){
				num_allpattern *= NC[e];
			}
			if(LD.size() / num_allpattern < 5){
				PC.remove(PC.indexOf(i));
				break;
			}
			double statistics = G2_statistics(i, NV - 1, pattern);
			double dimension = (NC[i] - 1) * (NC[NV - 1] - 1);
			for(int e: pattern){
				dimension *= NC[e];
			}
			if(dimension == 0) dimension = 1;
			double pvalue = 1-  new ChiSquaredDistribution(dimension).cumulativeProbability(statistics);
			if(pvalue > 0.05){
				PC.remove(PC.indexOf(i));
				return true;
			}
		}
		return false;
	}

	private double G2_statistics(int x, int y, ArrayList<Integer> z){
		HashMap<Integer, Integer> n_z = new HashMap<>();
		HashMap<Integer, int[]> n_xz = new HashMap<>();
		HashMap<Integer, int[]> n_yz = new HashMap<>();
		HashMap<Integer, int[][]> n_xyz = new HashMap<>();
		for(int l=0; l<LD.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int i = 0; i < NV; ++i) currentLearnData[i] = LD.get(l)[i];
			int hash = 0;
			for(int i = 0; i < z.size(); ++i) hash += Math.pow(MNC, i) * currentLearnData[z.get(i)];
			if(n_z.containsKey(hash)){
				n_z.put(hash, n_z.get(hash) + 1);
				(n_xz.get(hash)[currentLearnData[x]])++;
				//System.out.println(y + ", "+NC[y]);
				(n_yz.get(hash)[currentLearnData[y]])++;
				(n_xyz.get(hash)[currentLearnData[x]][currentLearnData[y]])++;
			}else{
				//n_z
				n_z.put(hash, 1);

				//n_xz
				int[] tx = new int[NC[x]];
				for(int j = 0; j < NC[x]; ++j){
					tx[j] = 0;
				}
				tx[currentLearnData[x]] = 1;
				n_xz.put(hash, tx);

				//n_yz
				int[] ty = new int[NC[y]];
				for(int j = 0; j < NC[y]; ++j){
					ty[j] = 0;
				}
				ty[currentLearnData[y]] = 1;
				n_yz.put(hash, ty);

				//n_xyz
				int[][] t = new int[NC[x]][NC[y]];
				for(int j = 0; j < NC[x]; ++j){
					for(int jj = 0; jj < NC[y]; ++jj){
						t[j][jj] = 0;
					}
				}
				t[currentLearnData[x]][currentLearnData[y]] = 1;
				n_xyz.put(hash, t);
			}
		}

		//G2
		double statistics = 0;
		for(Map.Entry<Integer, Integer> e : n_z.entrySet()) {
			int hash = e.getKey();
			double s_z = e.getValue();
			for(int i = 0; i < NC[x]; ++i){
				double s_xz = n_xz.get(hash)[i];
				for(int j = 0; j < NC[y]; ++j){
					double s_yz = n_yz.get(hash)[j];
					double s_xyz = n_xyz.get(hash)[i][j];
					if(s_xyz != 0){
						statistics += s_xyz * (Math.log(s_xyz) + Math.log(s_z) - Math.log(s_xz) - Math.log(s_yz));
					}
					//System.out.println(s_z+ ", "+s_xz+ ", "+s_yz+ ", "+s_xyz);
				}
			}
		}
		statistics *= 2;
		return statistics;
	}

	private ArrayList<Integer> DecodeParents(int j, ArrayList<Integer> candidate){
		ArrayList<Integer> parents = new ArrayList<>();
		for(int c = 0; c < candidate.size(); ++c){
			if((j & (1 << c)) != 0){
				parents.add(candidate.get(c));
			}
		}
		return parents;
	}

	private ArrayList<Integer> DecodeParents(int i, int j, ArrayList<Integer> candidate){
		ArrayList<Integer> parents = new ArrayList<>();
		for(int c = 0; c < candidate.size() - 1; ++c){
			if((j & (1 << c)) != 0){
				if(c < i){
					parents.add(candidate.get(c));
				}else{
					parents.add(candidate.get(c+1));
				}
			}
		}
		return parents;
	}

	private ArrayList<Integer> HITON_PC(){
		ArrayList<Integer> CPC = _HITON_PC();
		for(int i = 0; i < CPC.size(); ++i){
			//backup data
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[NV];
				for(int j = 0; j < NV; ++j){
					data[j] = LD.get(d)[j];
				}
				tLD.add(data);
			}
			int[] tNC = new int[NV];
			for(int j = 0; j < NV; ++j){
				tNC[j] = NC[j];
			}
			//learn attributes' PC
			arrangeClassVariable(CPC.get(i));
			ArrayList<Integer> attributesPC = _HITON_PC();
			if(!attributesPC.contains(NV - 2)){
				CPC.remove(i);
				--i;
			}
			LD = tLD;
			NC = tNC;
		}
		return CPC;
	}

	private ArrayList<Integer> _HITON_PC(){
		ArrayList<Integer> PC = new ArrayList<>();
		ArrayList<Integer> deplist_largeOrder = sortdep_largeOrder();
		for(int c = 0; c < deplist_largeOrder.size(); ++c){
			int i = deplist_largeOrder.get(c);
			PC.add(i);
			Collections.sort(PC);
			//remove false positive
			for(int pc = 0; pc < PC.size(); ++pc){
				boolean remove = remove_false_positive(PC, PC.get(pc));
				if(remove) --pc;
			}
		}
		return PC;
	}

	private ArrayList<Integer> sortdep_largeOrder(){
		ArrayList<Double> sortdep = new ArrayList<>();
		ArrayList<Integer> sortdepIndex = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i) sortdep.add(0.0);
		for(int i = 0; i < NV - 1; ++i) sortdepIndex.add(0);
		ArrayList<Integer> empty = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i){
			double statistics = G2_statistics(i, NV - 1, empty);
			double dimension = (NC[i] - 1)*(NC[NV - 1] - 1);
			if(dimension == 0) dimension = 1;
			double pvalue = 1 - new ChiSquaredDistribution(dimension).cumulativeProbability(statistics);
			double dep = -pvalue;
			for(int j = 0; j < NV - 1; ++j){
				if(j == i){
					sortdep.set(j, dep);
					sortdepIndex.set(j, i);
					break;
				}
				if(sortdep.get(j) < dep){
					for(int k = i; k > j; --k){
						sortdep.set(k, sortdep.get(k-1));
						sortdepIndex.set(k, sortdepIndex.get(k-1));
					}
					sortdep.set(j, dep);
					sortdepIndex.set(j, i);
					break;
				}
			}
		}
		return sortdepIndex;
	}

	private ArrayList<Integer> PCMB(){
		ArrayList<Integer> CPC = _PCMB();
		for(int i = 0; i < CPC.size(); ++i){
			//backup data
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[NV];
				for(int j = 0; j < NV; ++j){
					data[j] = LD.get(d)[j];
				}
				tLD.add(data);
			}
			int[] tNC = new int[NV];
			for(int j = 0; j < NV; ++j){
				tNC[j] = NC[j];
			}
			//learn attributes' PC
			arrangeClassVariable(CPC.get(i));
			ArrayList<Integer> attributesPC = _PCMB();
			if(!attributesPC.contains(NV - 2)){
				CPC.remove(i);
				--i;
			}
			LD = tLD;
			NC = tNC;
		}
		return CPC;
	}

	private ArrayList<Integer> _PCMB(){
		ArrayList<Integer> PC = new ArrayList<>();
		ArrayList<Integer> candidate = new ArrayList<>();
		for(int i = 0; i < NV - 1; ++i) candidate.add(i);
		// add true positive
		while(!candidate.isEmpty()){
			int variable_maxdep = 0;
			double maxdep = -Math.pow(10, 100);
			ArrayList<Integer> removedList = new ArrayList<>();
			for(int c = 0; c < candidate.size(); ++c){
				int i = candidate.get(c);
				if(PC.contains(i)) continue;
				double dep = getMindep(PC, i);
				if(-dep > 0.05){
					removedList.add(candidate.get(c));
				}
				if(dep > maxdep){
					maxdep = dep;
					variable_maxdep = i;
				}
			}
			for(int r: removedList){
				candidate.remove(candidate.indexOf(r));
			}
			PC.add(variable_maxdep);
			if(candidate.contains(variable_maxdep)){
				candidate.remove(candidate.indexOf(variable_maxdep));
			}
			Collections.sort(PC);
			//remove false positive
			for(int i = 0; i < PC.size(); ++i){
				boolean remove = remove_false_positive(PC, PC.get(i));
				if(remove) --i;
			}
		}
		return PC;
	}

	private void arrangeClassVariable(int c){
		for(int d = 0; d < LD.size(); ++d){
			int t = LD.get(d)[c];
			for(int i = c; i < NV - 1; ++i){
				LD.get(d)[i] = LD.get(d)[i+1];
			}
			LD.get(d)[NV - 1] = t;
		}
		int t = NC[c];
		for(int i = c; i < NV - 1; ++i){
			NC[i] = NC[i+1];
		}
		NC[NV - 1] = t;
	}

}
