import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Queue;
import java.util.Random;
import java.util.Set;


public class MCDAGGES {
	public static int NV;
	public static int[] NC;
	public static int MNC;
	public static ArrayList<int[]> LD;

	public static int COUNT;
	public static int CrossValidation;
	public ArrayList<int[]> RD;
	public int fANB;
	public ArrayList<ArrayList<Integer>> structure;
	public double[][][][] scoreList_GES;
	public int end_condition = 10000000;
	public ArrayList<double[]> JPL; //joint probability list

	public static void main(String[] args) throws IOException, InterruptedException {
		new MCDAGGES("data/Balance.csv", "data", 10);
	}

	public MCDAGGES(String path_to_dataset, String path_to_directory_for_output, int cv) throws IOException, InterruptedException{
		System.out.print("Exeute MCDAGGES with ");
		System.out.println(cv + "-fold cross-validation.");
		CrossValidation = cv;
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);
			int[][] str = new int[NV][NV];
			for(int i = 0; i < NV; ++i){
				for(int k = 0; k < NV; ++k){
					if(i == k) continue;
					if(i == NV - 1 || k == NV - 1){
						str[i][k] = 2;
					}
				}
			}
			structure = new ArrayList<>();
			for(int i = 0; i < NV; ++i){
				structure.add(new ArrayList<Integer>());
				if(i != NV - 1){
					structure.get(i).add(NV - 1);
				}
			}
			initJPL(structure);
			ArrayList<HashMap<BigInteger, int[]>> n1 = makeCft(structure);
			double s = calcCLL(structure, n1);
			while(true){
				s = LearningMCDAGGES_parallel(str, s);
				if(s == end_condition) break;
			}
			PDAGtoDAG_forMCDAG(str);
			structure = new ArrayList<>();
			for(int i = 0; i < NV; ++i){
				structure.add(new ArrayList<Integer>());
				for(int k = 0; k < NV; ++k){
					if(str[k][i] == 1){
						structure.get(i).add(k);
					}
				}
			}
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	public class GD implements Runnable {
		private double[][][][] scorelist_GES;
		private ArrayList<ArrayList<Integer>> str;
		private int p1;
		private int p2;
		private int type;
		private int[][] str_array;
		public GD(double[][][][] scorelist_GES, int[][] str_array, int type, int p1, int p2) { // p1->p2
			this.scorelist_GES = scorelist_GES;
			this.str_array = str_array;
			this.p1 = p1;
			this.p2 = p2;
			this.type = type;
		}
		@Override
		public void run() {
			if(type == 0){
				validInsert(str_array, type, p1, p2);
			}else{
				validDelete(str_array, type, p1, p2);
			}
		}
	}

	void scoreListInit(){
		scoreList_GES = new double[3][NV][NV][NV + 1];
		for(int i = 0; i < 3; ++i){
			for(int j = 0; j < NV; ++j){
				for(int k = 0; k < NV; ++k){
					for(int l = 0; l < NV + 1; ++l){
						scoreList_GES[i][j][k][l] = -1000000000;
					}
				}
			}
		}
	}

	protected void readData(String DataBaseName) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(DataBaseName)));
		ArrayList<String[]> tRD = new ArrayList<String[]>();
		String line = br.readLine();
		NV = line.split(",").length;
		while(line != null){
			String[] data = line.split(",").clone();
			tRD.add(data);
			line = br.readLine();
		}

		int[] max = new int[NV];
		RD = new ArrayList<int[]>();
		for(int i = 0; i < tRD.size(); ++i){
			RD.add(new int[NV]);
			for(int j = 0; j < NV; ++j){
				RD.get(i)[j] = Integer.parseInt(tRD.get(i)[j]);
				if(max[j] < RD.get(i)[j]){
					max[j] = RD.get(i)[j];
				}
			}
		}
		MNC = 0;
		NC = new int[NV];
		for(int i = 0; i < NV; ++i){
			NC[i] = max[i] + 1;
			if(MNC < NC[i]){
				MNC = NC[i];
			}
		}
		br.close();
	}

	protected void readLearningData(int j) throws IOException{
		if(CrossValidation == 1){
			LD = RD;
			return;
		}
		LD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = 0; i < RD.size(); ++i){
				if(i < j*(RD.size()/CrossValidation) || (j+1)*(RD.size()/CrossValidation) <= i)
					LD.add(RD.get(i));
			}
		}
		else{
			for(int i = 0; i < (CrossValidation - 1)*(RD.size()/CrossValidation); ++i){
				LD.add(RD.get(i));
			}
		}
	}

	public void writeStructure(String path, int j){
		try{
			File outputDir = new File(path);
			if (outputDir.exists()){
				File file = new File(path + "/" + (j+1) + ".txt");
				file.createNewFile();
				PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
				for(int i = 0; i < NV; ++i){
					pw.print(i);
					pw.print("\t");
				}
				pw.println();
				for(int i = 0; i < NV; ++i){
					ArrayList<Integer> t = structure.get(i);
					if(t.size() != 0){
						pw.print(t.get(0));
						for(int p = 1; p < t.size(); ++p){
							pw.print(",");
							pw.print(t.get(p));
						}
					}else{
						pw.print(-1);
					}
					pw.print("\t");
				}
				pw.close();
			}else{
				System.out.println();
				System.out.println("The output directory not found.");
				System.exit(0);
			}
		}catch(IOException e){
			System.out.println(e);
		}
	}

	private double LearningMCDAGGES_parallel(int[][] str, double score) throws InterruptedException{ // greedy search
		scoreListInit();


		ArrayList<Thread> thrList = new ArrayList<>();
		for(int i = 0; i < NV - fANB; ++i){
			//add
			//System.out.println("add");
			for(int j = 0; j < NV - fANB; ++j){
				if(i == j) continue;
				if(str[i][j] != 0 || str[j][i] != 0) continue;
				if(!(j == NV - 1 || str[NV - 1][i] == 1 || str[NV - 1][i] == 2)) continue;
				Runnable runnable = new GD(scoreList_GES, str, 0, j, i);
				Thread thr = new Thread(runnable);
				thr.start();
				thrList.add(thr);
			}
			//delete
			//System.out.println("delete");
			for(int j = 0; j < NV - fANB; ++j){
				if(i == j) continue;
				if(str[j][i] == 0) continue;
				Runnable runnable = new GD(scoreList_GES, str, 1, j, i);
				Thread thr = new Thread(runnable);
				thr.start();
				thrList.add(thr);
			}
		}

		for(int c = 0; c < thrList.size(); ++c){
			thrList.get(c).join();
		}



		double max = score;
		int type = -1;
		int p1 = 0;
		int p2 = 0;
		int index = 0;
		for(int i = 0; i < 3; ++i){
			for(int j = 0; j < NV; ++j){
				for(int k = 0; k < NV; ++k){
					for(int l = 0; l < NV + 1; ++l){
						//if(scoreList_GES[i][j][k][l] != -1000000000){
						//	System.out.println(i+","+j+","+k+","+l+","+scoreList_GES[i][j][k][l]);
						//}
						if(max < scoreList_GES[i][j][k][l]){
							max = scoreList_GES[i][j][k][l];
							type = i;
							p2 = j;
							p1 = k;
							index = l;
						}
					}
				}
			}
		}

		switch(type){
		case 0:
			updateJPL_Insert(str, p1, 2, index);
			str[p1][p2] = 1;
			if(index != NV){
				str[index][p2] = 1;
				str[p2][index] = 0;
			}
			break;
		case 1:
			updateJPL_Delete(str, p1, 2, index);
			str[p1][p2] = 0;
			str[p2][p1] = 0;
			if(index != NV){
				str[index][p2] = 0;
				str[p2][index] = 1;
				if(str[p1][index] == 2){
					str[index][p1] = 0;
					str[p1][index] = 1;
				}
			}
			//post-deletȅ
			if(p1 == NV - 1){
				for(int i = 0; i < NV; ++i){
					if(str[i][p2] == 1){
						str[i][p2] = 0;
					}else if(str[i][p2] == 2){
						str[i][p2] = 0;
						str[p2][i] = 1;
					}
				}
			}
			break;
		case -1:
			max = 10000000;
			break;
		default: System.out.println(); System.out.print("ERROR"); System.exit(0);
		}

		PDAGtoDAG_forMCDAG(str);
		DAGtoCPDAG(str);
		return max;
	}

	private void initJPL(ArrayList<ArrayList<Integer>> parents){
		JPL = new ArrayList<>();
		for(int d = 0; d < LD.size(); ++d){
			JPL.add(new double[NC[NV - 1]]);
		}
		ArrayList<HashMap<BigInteger, int[]>> n = makeCft(parents);
		for(int d = 0; d < LD.size(); ++d){
			int[] data = LD.get(d).clone();
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				double jp = 0; //joint probability
				for(int i = 0; i < NV; ++i){
					BigInteger hash = makeKey(data, i, parents);
					double n_ij = 0;
					double n_ijk = 0;
					if(n.get(i).containsKey(hash)){
						for(int k = 0; k < NC[i]; ++k){
							n_ij += n.get(i).get(hash)[k];
						}
						n_ijk = n.get(i).get(hash)[data[i]];
					}
					jp += Math.log(n_ijk + 1) - Math.log(n_ij + NC[i]);
				}
				JPL.get(d)[c] = jp;
			}
		}
	}

	private void validInsert(int[][] str, int type, int p1, int p2){
		ArrayList<Integer> NAyx = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			if(str[p2][i] == 2 && (str[i][p1] == 1 || str[p1][i] == 1 || str[p1][i] == 2)){
				NAyx.add(i);
			}
		}
		if(!checkClique(NAyx, str)) return; //T = ӂɊւValidInsert̏1

		boolean condition2_for_valid = !checkExistSemiDirectedPathWithoutBlock(str, p1, p2, NAyx);
		if(condition2_for_valid){
			Insert(str, type, p1, p2, NV);
		}

		for(int T = 0; T < NV; ++T){
			if(T == p1 || T == p2) continue;
			if(str[T][p2] == 2 && str[p1][T] == 0 && str[p1][T] == 0) continue;//TYneighborXnot adjacentȂƂȂD
			ArrayList<Integer> NAyxT = (ArrayList<Integer>) NAyx.clone();
			NAyxT.add(T);
			if(condition2_for_valid || !checkExistSemiDirectedPathWithoutBlock(str, p1, p2, NAyxT)){//̏ŏ\C̏trueȂ悢
				Insert(str, type, p1, p2, T);
			}
		}
	}

	private void Insert(int[][] str, int type, int x, int y, int T){
		ArrayList<Integer> g = new ArrayList<>();
		ArrayList<Integer> gs = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			if(str[i][y] == 1 || i == T){
				g.add(i);
				gs.add(i);
			}
		}
		gs.add(x);
		Collections.sort(gs);
		if(!gs.contains(NV - 1)){
			g.add(NV - 1);
			gs.add(NV - 1);
		}
		HashMap<BigInteger, int[]> n_g = makeCft_i(y, g);
		HashMap<BigInteger, int[]> n_gs = makeCft_i(y, gs);
		double CLL = 0;
		for(int d = 0; d < LD.size(); ++d){
			int[] data = LD.get(d).clone();
			double sum = 0;
			int save = data[NV - 1];
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				BigInteger hash_g = makeKey_i(data, g);
				BigInteger hash_gs = makeKey_i(data, gs);
				double nij_g = 0;
				double nij_gs = 0;
				double nijk_g = 0;
				double nijk_gs = 0;
				if(n_g.containsKey(hash_g)){
					for(int k = 0; k < NC[y]; ++k){
						nij_g += n_g.get(hash_g)[k];
					}
					nijk_g = n_g.get(hash_g)[data[y]];
				}
				if(n_gs.containsKey(hash_gs)){
					for(int k = 0; k < NC[y]; ++k){
						nij_gs += n_gs.get(hash_gs)[k];
					}
					nijk_gs = n_gs.get(hash_gs)[data[y]];
				}
				double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
				double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
				sum += Math.exp(JPL.get(d)[c] + p_gs - p_g);
				//System.out.println(p_g + ",,," + p_gs);
			}
			data[NV - 1] = save;
			BigInteger hash_g = makeKey_i(data, g);
			BigInteger hash_gs = makeKey_i(data, gs);
			double nij_g = 0;
			double nij_gs = 0;
			double nijk_g = 0;
			double nijk_gs = 0;
			if(n_g.containsKey(hash_g)){
				for(int k = 0; k < NC[y]; ++k){
					nij_g += n_g.get(hash_g)[k];
				}
				nijk_g = n_g.get(hash_g)[data[y]];
			}
			if(n_gs.containsKey(hash_gs)){
				for(int k = 0; k < NC[y]; ++k){
					nij_gs += n_gs.get(hash_gs)[k];
				}
				nijk_gs = n_gs.get(hash_gs)[data[y]];
			}
			double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
			double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
			CLL += (JPL.get(d)[save] + p_gs - p_g) - Math.log(sum);
		}
		scoreList_GES[type][x][y][T] = CLL;
	}

	private void validDelete(int[][] str, int type, int p1, int p2){
		ArrayList<Integer> NAyx = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			if(str[p2][i] == 2 && (str[i][p1] == 1 || str[p1][i] == 1 || str[p1][i] == 2)){
				NAyx.add(i);
			}
		}
		boolean condition_for_valid = checkClique(NAyx, str);//H = ӂɊւValidDeletȅ
		if(condition_for_valid){
			Delete(str, type, p1, p2, NV);
		}

		for(int H: NAyx){
			ArrayList<Integer> NAyxH = (ArrayList<Integer>) NAyx.clone();
			NAyxH.remove(NAyx.indexOf(H));
			if(condition_for_valid || checkClique(NAyxH, str)){//̏ŏ\C̏trueȂ悢
				Delete(str, type, p1, p2, H);
			}
		}
	}

	private void Delete(int[][] str, int type, int x, int y, int H){
		ArrayList<Integer> g = new ArrayList<>();
		ArrayList<Integer> gs = new ArrayList<>();
		for(int i = 0; i < NV; ++i){ //H͖ړIϐłĂ͂ȂȂD
			if(str[i][y] == 1 && i != H){
				if(x != i){
					g.add(i);
					gs.add(i);
				}
			}
		}
		g.add(x);
		Collections.sort(g);
		if(!g.contains(NV - 1)){
			g.add(NV - 1);
			gs.add(NV - 1);
		}
		HashMap<BigInteger, int[]> n_g = makeCft_i(y, g);
		HashMap<BigInteger, int[]> n_gs = makeCft_i(y, gs);
		double CLL = 0;
		for(int d = 0; d < LD.size(); ++d){
			int[] data = LD.get(d).clone();
			double sum = 0;
			int save = data[NV - 1];
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				BigInteger hash_g = makeKey_i(data, g);
				BigInteger hash_gs = makeKey_i(data, gs);
				double nij_g = 0;
				double nij_gs = 0;
				double nijk_g = 0;
				double nijk_gs = 0;
				if(n_g.containsKey(hash_g)){
					for(int k = 0; k < NC[y]; ++k){
						nij_g += n_g.get(hash_g)[k];
					}
					nijk_g = n_g.get(hash_g)[data[y]];
				}
				if(n_gs.containsKey(hash_gs)){
					for(int k = 0; k < NC[y]; ++k){
						nij_gs += n_gs.get(hash_gs)[k];
					}
					nijk_gs = n_gs.get(hash_gs)[data[y]];
				}
				double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
				double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
				sum += Math.exp(JPL.get(d)[c] + p_gs - p_g);
			}
			data[NV - 1] = save;
			BigInteger hash_g = makeKey_i(data, g);
			BigInteger hash_gs = makeKey_i(data, gs);
			double nij_g = 0;
			double nij_gs = 0;
			double nijk_g = 0;
			double nijk_gs = 0;
			if(n_g.containsKey(hash_g)){
				for(int k = 0; k < NC[y]; ++k){
					nij_g += n_g.get(hash_g)[k];
				}
				nijk_g = n_g.get(hash_g)[data[y]];
			}
			if(n_gs.containsKey(hash_gs)){
				for(int k = 0; k < NC[y]; ++k){
					nij_gs += n_gs.get(hash_gs)[k];
				}
				nijk_gs = n_gs.get(hash_gs)[data[y]];
			}
			double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
			double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
			CLL += (JPL.get(d)[save] + p_gs - p_g) - Math.log(sum);
		}
		scoreList_GES[type][x][y][H] = CLL;
	}

	private boolean checkClique(ArrayList<Integer> clq, int[][] str){
		for(int i = 0; i < clq.size() - 1; ++i){
			int e1 = clq.get(i);
			for(int j = i + 1; j < clq.size(); ++j){
				int e2 = clq.get(j);
				if(str[e1][e2] == 0 && str[e2][e1] == 0) return false;
			}
		}
		return true;
	}

	private boolean checkExistSemiDirectedPathWithoutBlock(int[][] str, int x, int y, ArrayList<Integer> block){ //yxւpathblockɂăubNȂǂ݂D
		Set<Integer> mile = new HashSet();
		Queue<Integer> candidate = new ArrayDeque();
		candidate.add(y);
		while(!candidate.isEmpty()){
			int now = candidate.poll();
			if(now == x){
				return true;
			}
			for(int i = 0; i < NV; ++i){
				if(str[now][i] != 0 && !mile.contains(i) && !block.contains(i)){
					candidate.add(i);
				}
			}
			mile.add(now);
		}
		return false;
	}



	private void updateJPL_Insert(int[][] str, int x, int y, int T){
		ArrayList<Integer> g = new ArrayList<>();
		ArrayList<Integer> gs = new ArrayList<>();
		for(int i = 0; i < NV; ++i){
			if(str[i][y] == 1 || i == T){
				g.add(i);
				gs.add(i);
			}
		}
		gs.add(x);
		Collections.sort(gs);
		if(!gs.contains(NV - 1)){
			g.add(NV - 1);
			gs.add(NV - 1);
		}
		HashMap<BigInteger, int[]> n_g = makeCft_i(y, g);
		HashMap<BigInteger, int[]> n_gs = makeCft_i(y, gs);
		for(int d = 0; d < LD.size(); ++d){
			int[] data = LD.get(d).clone();
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				BigInteger hash_g = makeKey_i(data, g);
				BigInteger hash_gs = makeKey_i(data, gs);
				double nij_g = 0;
				double nij_gs = 0;
				double nijk_g = 0;
				double nijk_gs = 0;
				if(n_g.containsKey(hash_g)){
					for(int k = 0; k < NC[y]; ++k){
						nij_g += n_g.get(hash_g)[k];
					}
					nijk_g = n_g.get(hash_g)[data[y]];
				}
				if(n_gs.containsKey(hash_gs)){
					for(int k = 0; k < NC[y]; ++k){
						nij_gs += n_gs.get(hash_gs)[k];
					}
					nijk_gs = n_gs.get(hash_gs)[data[y]];
				}
				double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
				double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
				JPL.get(d)[c] += p_gs - p_g;
			}
		}
	}

	private void updateJPL_Delete(int[][] str, int x, int y, int H){
		ArrayList<Integer> g = new ArrayList<>();
		ArrayList<Integer> gs = new ArrayList<>();
		for(int i = 0; i < NV; ++i){ //H͖ړIϐłĂ͂ȂȂD
			if(str[i][y] == 1 && i != H){
				if(x != i){
					g.add(i);
					gs.add(i);
				}
			}
		}
		g.add(x);
		Collections.sort(g);
		if(!g.contains(NV - 1)){
			g.add(NV - 1);
			gs.add(NV - 1);
		}
		HashMap<BigInteger, int[]> n_g = makeCft_i(y, g);
		HashMap<BigInteger, int[]> n_gs = makeCft_i(y, gs);
		for(int d = 0; d < LD.size(); ++d){
			int[] data = LD.get(d).clone();
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				BigInteger hash_g = makeKey_i(data, g);
				BigInteger hash_gs = makeKey_i(data, gs);
				double nij_g = 0;
				double nij_gs = 0;
				double nijk_g = 0;
				double nijk_gs = 0;
				if(n_g.containsKey(hash_g)){
					for(int k = 0; k < NC[y]; ++k){
						nij_g += n_g.get(hash_g)[k];
					}
					nijk_g = n_g.get(hash_g)[data[y]];
				}
				if(n_gs.containsKey(hash_gs)){
					for(int k = 0; k < NC[y]; ++k){
						nij_gs += n_gs.get(hash_gs)[k];
					}
					nijk_gs = n_gs.get(hash_gs)[data[y]];
				}
				double p_g = Math.log(nijk_g + 1) - Math.log(nij_g + NC[y]);
				double p_gs = Math.log(nijk_gs + 1) - Math.log(nij_gs + NC[y]);
				JPL.get(d)[c] += p_gs - p_g;
			}
		}
	}

	private void PDAGtoDAG_forMCDAG(int[][] str){//ӁFMCDAGp
		Set<Integer> mile = new HashSet();
		Queue<Integer> candidate = new ArrayDeque();
		candidate.add(NV - 1);
		while(!candidate.isEmpty()){
			int now = candidate.poll();
			for(int i = 0; i < NV; ++i){
				if(str[now][i] != 0 && !mile.contains(i)){
					str[now][i] = 1;
					str[i][now] = 0;
					candidate.add(i);
				}
			}
			mile.add(now);
		}
		for(int i = 0; i < NV; ++i){
			if(!mile.contains(i)){
				for(int j = 0; j < NV; ++j){
					if(str[i][j] == 2){
						str[i][j] = 1;
						str[j][i] = 0;
					}
				}
			}
		}
	}

	private void DAGtoCPDAG(int[][] str){
		Set<Integer> mile = new HashSet();
		Queue<Integer> candidate = new ArrayDeque();
		// ׂĂrootm[hD
		for(int i = 0; i < NV; ++i){
			boolean leaf = true;
			for(int j = 0; j < NV; ++j){
				if(str[j][i] == 1){
					leaf = false;
					break;
				}
			}
			if(leaf){
				candidate.add(i);
			}
		}
		while(!candidate.isEmpty()){
			int now = candidate.poll();
			mile.add(now);
			ArrayList<Integer> parents_now = new ArrayList<>();
			for(int i = 0; i < NV; ++i){
				if(str[i][now] == 1){
					parents_now.add(i);
				}
			}
			for(int i = 0; i < NV; ++i){
				if(str[now][i] == 1 && !mile.contains(i)){
					candidate.add(i);
					boolean undirect = true;
					for(int pa : parents_now){
						if(str[pa][i] == 0 && str[i][pa] == 0){
							undirect = false;
						}
					}
					ArrayList<Integer> parents_i = new ArrayList<>();
					for(int j = 0; j < NV; ++j){
						if(j != now && str[j][i] == 1){
							parents_i.add(j);
						}
					}
					for(int pa : parents_i){
						if(str[pa][i] == 0 && str[i][pa] == 0){
							undirect = false;
						}
					}
					if(undirect){
						str[now][i] = 2;
						str[i][now] = 2;
					}
				}
			}
		}
	}

	private HashMap<BigInteger, int[]> makeCft_i(int i, ArrayList<Integer> parents){
		HashMap<BigInteger, int[]> n_i = new HashMap<>();
		for(int l=0; l<LD.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int c = 0; c < NV; ++c){
				currentLearnData[c] = LD.get(l)[c];
			}
			//eԂ𐔒lɕϊ
			BigInteger hash = makeKey_i(currentLearnData, parents);
			if(n_i.containsKey(hash)){ //̐eԂmapɊɓo^ĂȂA
				(n_i.get(hash)[currentLearnData[i]])++; //lXV
			}else{ // eԂCFTɓo^ĂȂA
				// VKɍ쐬zmapɓo^
				int[] t = new int[NC[i]];
				for(int j = 0; j < NC[i]; ++j){
					t[j] = 0;
				}
				t[currentLearnData[i]] = 1;
				n_i.put(hash, t);
			}
		}
		return n_i;
	}

	private BigInteger makeKey_i(int[] data, ArrayList<Integer> parentList){ // e -> l ɕϊ
		if(!parentList.isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parentList.size(); ++p){
				int t = parentList.get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	private double calcCLL(ArrayList<ArrayList<Integer>> parents, ArrayList<HashMap<BigInteger, int[]>> n){
		double L = 0.0;
		for(int d = 0; d < LD.size(); ++d){
			int save = LD.get(d)[NV - 1];
			double[][] p = new double[NV][NC[NV - 1]];
			int[] data = new int[NV];
			for(int i = 0; i < NV; ++i){
				data[i] = LD.get(d)[i];
			}
			for(int c = 0; c < NC[NV - 1]; ++c){
				data[NV - 1] = c;
				for(int i = 0; i < NV; ++i){
					BigInteger key = makeKey(data, i, parents);
					if(n.get(i).containsKey(key)){
						double n_ij = 0;
						for(int k = 0; k < NC[i]; ++k){
							n_ij += n.get(i).get(key)[k];
						}
						double n_ijk = n.get(i).get(key)[data[i]];
						p[i][c] = n_ijk / n_ij;
					}else{
						p[i][c] = 0;
					}
				}
			}
			double total = 0.0;
			double[] lc = new double[NC[NV - 1]];
			for(int v = 0; v < NC[NV - 1]; ++v){
				boolean pzero = false;
				lc[v] = 0.0;
				for(int i = 0; i < NV; ++i){
					if(p[i][v] != 0){
						lc[v] += Math.log(p[i][v]);
					}else{
						pzero = true;
						break;
					}
				}
				if(!pzero){
					total += Math.exp(lc[v]);
				}
			}
			L += lc[save] - Math.log(total);
		}

		double[] denomi = new double[NV];
		for(int i = 0; i < NV; ++i){
			double denominator = 1.0;
			for(int j = 0; j < parents.get(i).size(); ++j){
				denominator *= NC[parents.get(i).get(j)];
			}
			denomi[i] = denominator;
		}
		return L;
	}

	private ArrayList<HashMap<BigInteger, int[]>> makeCft(ArrayList<ArrayList<Integer>> parents){
		ArrayList<HashMap<BigInteger, int[]>> n = new ArrayList<HashMap<BigInteger, int[]>>();
		for(int i = 0; i < NV; ++i) n.add(new HashMap<BigInteger, int[]>());
		for(int l=0; l<LD.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int i = 0; i < NV; ++i) currentLearnData[i] = LD.get(l)[i];
			for(int i = 0; i < NV; i++){
				BigInteger hash = makeKey(currentLearnData, i, parents);
				if(n.get(i).containsKey(hash)){
					(n.get(i).get(hash)[currentLearnData[i]])++;
				}else{
					int[] t = new int[NC[i]];
					for(int j = 0; j < NC[i]; ++j){
						t[j] = 0;
					}
					t[currentLearnData[i]] = 1;
					n.get(i).put(hash, t);
				}
			}
		}
		return n;
	}

	private BigInteger makeKey(int[] data, int i, ArrayList<ArrayList<Integer>> parentList){
		if(!parentList.get(i).isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parentList.get(i).size(); ++p){
				int t = parentList.get(i).get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	private ArrayList<ArrayList<Integer>> makingStructure(int anb){
		ArrayList<ArrayList<Integer>> str = new ArrayList<>();
		for(int i = 0; i < NV - anb; ++i){
			str.add(new ArrayList<Integer>());
		}
		ArrayList<Integer> order = randomOrder(NV - anb);

		Random rnd = new Random();
		for(int i = 0; i < NV - anb; ++i){
			for(int j = 0; j < i; ++j){
				int f = rnd.nextInt(2);
				if(f == 0){
					str.get(order.get(i)).add(order.get(j));
				}
			}
			Collections.sort(str.get(order.get(i)));
		}
		if(anb == 1){
			for(int i = 0; i < NV - 1; ++i){
				str.get(i).add(NV - 1);
			}
			str.add(new ArrayList<Integer>());
		}
		return str;
	}

	private ArrayList<Integer> randomOrder(int size){
		ArrayList<Integer> r = new ArrayList<>();
		for(int i = 0; i < size; ++i){
			r.add(i);
		}
		Collections.shuffle(r);
		return r;
	}

	private BigInteger product(int a, int b){
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}
}
