import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

public class fsANB_BDeu {
	public int fANB; //learning ANB if fABN = 1, learning GBN otherwise.
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public int CrossValidation; //How many times split data for cross validation.
	public ArrayList<int[]> RD; //Read data
	public ArrayList<int[]> LD; //Learning data for each validation.
	public getLocalScore gls;
	public getBestParents gbp;
	public getBestNets gbn;
	public ArrayList<ArrayList<Integer>> structure;
	public ArrayList<double[]> prob;

	public fsANB_BDeu(String path_to_dataset, String path_to_directory_for_output, int cv, int fsMethod) throws IOException{
		System.out.print("Exeute ANB-BDeu with ");
		System.out.print(cv + "-fold cross-validation ");
		System.out.print("using feature selection ");
		switch(fsMethod){
		case 2: System.out.println("S2TMB."); break;
		case 3: System.out.println("MMPC."); break;
		case 4: System.out.println("HITON-PC."); break;
		case 5: System.out.println("PCMB."); break;
		default: System.out.println("Feature selection number is invalid."); System.exit(0); break;
		}
		CrossValidation = cv;
		fANB = 1;
		double alpha = 1.0;// equivalent sample size
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));

			readLearningData(j);

			//search MB
			FeatureSelection fs = new FeatureSelection(fsMethod, NV, NC, MNC, LD);
			ArrayList<Integer> MB = fs.fset;
			MB.add(NV - 1);

			//make data for MB
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[MB.size()];
				for(int i = 0; i < MB.size(); ++i){
					data[i] = LD.get(d)[MB.get(i)];
				}
				tLD.add(data);
			}
			int[] tNC = new int[MB.size()];
			for(int i = 0; i < MB.size(); ++i){
				tNC[i] = NC[MB.get(i)];
			}

			int saveNV = NV;
			int[] saveNC = NC.clone();
			NV = MB.size();
			LD = tLD;
			NC = tNC;

			gls = new getLocalScore(fANB, LD, NV, NC, MNC, alpha);
			gbp = new getBestParents(fANB, NV, NC, MNC, gls.LS, false);
			gbn = new getBestNets(fANB, NV, NC, MNC, gls.LS, gbp.BP);
			ArrayList<ArrayList<Integer>> tstructure = gbn.learnedStructure;
			structure = new ArrayList<>();
			for(int i = 0; i < saveNV; ++i) structure.add(new ArrayList<>());
			for(int i = 0; i < tstructure.size(); ++i){
				for(int pa = 0; pa < tstructure.get(i).size(); ++pa){
					structure.get(MB.get(i)).add(MB.get(tstructure.get(i).get(pa)));
				}
			}
			NV = saveNV;
			NC = saveNC;
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	public fsANB_BDeu(String path_to_dataset, String path_to_directory_for_output, int cv, double delta, double ess) throws IOException{
		System.out.print("Exeute fsANB-BDeu with delta=");
		System.out.print(delta + " and ess=" + ess);
		System.out.println(" using " + cv + "-fold cross-validation.");
		CrossValidation = cv;
		fANB = 1;
		double alpha = 1.0;// equivalent sample size
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);

			//search MB
			FeatureSelection fs = new FeatureSelection(delta, ess, NV, NC, MNC, LD);
			ArrayList<Integer> MB = fs.fset;
			MB.add(NV - 1);

			//make data for MB
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[MB.size()];
				for(int i = 0; i < MB.size(); ++i){
					data[i] = LD.get(d)[MB.get(i)];
				}
				tLD.add(data);
			}
			int[] tNC = new int[MB.size()];
			for(int i = 0; i < MB.size(); ++i){
				tNC[i] = NC[MB.get(i)];
			}

			int saveNV = NV;
			int[] saveNC = NC.clone();
			NV = MB.size();
			LD = tLD;
			NC = tNC;

			gls = new getLocalScore(fANB, LD, NV, NC, MNC, alpha);
			gbp = new getBestParents(fANB, NV, NC, MNC, gls.LS, false);
			gbn = new getBestNets(fANB, NV, NC, MNC, gls.LS, gbp.BP);
			ArrayList<ArrayList<Integer>> tstructure = gbn.learnedStructure;
			structure = new ArrayList<>();
			for(int i = 0; i < saveNV; ++i) structure.add(new ArrayList<>());
			for(int i = 0; i < tstructure.size(); ++i){
				for(int pa = 0; pa < tstructure.get(i).size(); ++pa){
					structure.get(MB.get(i)).add(MB.get(tstructure.get(i).get(pa)));
				}
			}
			NV = saveNV;
			NC = saveNC;
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	public fsANB_BDeu(String path_to_dataset, String path_to_directory_for_output, int cv) throws IOException{
		System.out.print("Exeute ANB-BDeu with ");
		System.out.println(cv + "-fold cross-validation while tuning delta and ess.");
		CrossValidation = cv;
		fANB = 1;
		double alpha = 1.0;// equivalent sample size
		readData(path_to_dataset);
		for(int j = 0; j < CrossValidation; ++j){
			System.out.print("Executing for cross validation "+(j+1));
			readLearningData(j);

			//decide threshold
			ArrayList<int[]> saveLD = LD;
			double[] candidate_Thv = {3, 20, 150};
			double[] candidate_Ess = {1, 2, 5};
			ArrayList<int[]> ld = new ArrayList<>();
			ArrayList<int[]> td = new ArrayList<>();
			for(int d = 0; d < LD.size() / 2; ++d){
				ld.add(LD.get(d).clone());
			}
			for(int d = LD.size() / 2; d < LD.size(); ++d){
				td.add(LD.get(d).clone());
			}
			double[][] correct = new double[candidate_Thv.length][candidate_Ess.length];
			for(int jj = 0; jj < 2; ++jj, LD = ld, ld = td, td = LD){
				for(int th = 0; th < candidate_Thv.length; ++th){
					for(int es = 0; es < candidate_Ess.length; ++es){
						double thv = candidate_Thv[th];
						double ess = candidate_Ess[es];
						//make MB
						LD = ld;
						FeatureSelection fs = new FeatureSelection(thv, ess, NV, NC, MNC, LD);
						ArrayList<Integer> MB = fs.fset;
						MB.add(NV - 1);
						//make data for MB
						ArrayList<int[]> tLD = new ArrayList<>();
						for(int d = 0; d < LD.size(); ++d){
							int[] data = new int[MB.size()];
							for(int i = 0; i < MB.size(); ++i){
								data[i] = LD.get(d)[MB.get(i)];
							}
							tLD.add(data);
						}
						int[] tNC = new int[MB.size()];
						for(int i = 0; i < MB.size(); ++i){
							tNC[i] = NC[MB.get(i)];
						}

						int saveNV = NV;
						int[] saveNC = NC.clone();
						NV = MB.size();
						LD = tLD;
						NC = tNC;

						gls = new getLocalScore(fANB, LD, NV, NC, MNC, alpha);
						gbp = new getBestParents(fANB, NV, NC, MNC, gls.LS, false);
						gbn = new getBestNets(fANB, NV, NC, MNC, gls.LS, gbp.BP);
						ArrayList<ArrayList<Integer>> tstructure = gbn.learnedStructure;
						structure = new ArrayList<>();
						for(int i = 0; i < saveNV; ++i) structure.add(new ArrayList<>());
						for(int i = 0; i < tstructure.size(); ++i){
							for(int pa = 0; pa < tstructure.get(i).size(); ++pa){
								structure.get(MB.get(i)).add(MB.get(tstructure.get(i).get(pa)));
							}
						}

						NV = saveNV;
						NC = saveNC;
						LD = ld;

						Classification(ld, td, alpha, structure);
						correct[th][es] += calcCorrect(td);
					}
				}
			}
			LD = saveLD;

			double max = correct[0][0];
			double thv = candidate_Thv[0];
			double ess = candidate_Ess[0];
			for(int th = 0; th < candidate_Thv.length; ++th){
				for(int es = 0; es < candidate_Ess.length; ++es){
					if(max < correct[th][es]){
						max = correct[th][es];
						thv = candidate_Thv[th];
						ess = candidate_Ess[es];
					}
				}
			}

			//search MB
			FeatureSelection fs = new FeatureSelection(thv, ess, NV, NC, MNC, LD);
			ArrayList<Integer> MB = fs.fset;
			MB.add(NV - 1);

			//make data for MB
			ArrayList<int[]> tLD = new ArrayList<>();
			for(int d = 0; d < LD.size(); ++d){
				int[] data = new int[MB.size()];
				for(int i = 0; i < MB.size(); ++i){
					data[i] = LD.get(d)[MB.get(i)];
				}
				tLD.add(data);
			}
			int[] tNC = new int[MB.size()];
			for(int i = 0; i < MB.size(); ++i){
				tNC[i] = NC[MB.get(i)];
			}

			int saveNV = NV;
			int[] saveNC = NC.clone();
			NV = MB.size();
			LD = tLD;
			NC = tNC;

			gls = new getLocalScore(fANB, LD, NV, NC, MNC, alpha);
			gbp = new getBestParents(fANB, NV, NC, MNC, gls.LS, false);
			gbn = new getBestNets(fANB, NV, NC, MNC, gls.LS, gbp.BP);
			ArrayList<ArrayList<Integer>> tstructure = gbn.learnedStructure;
			structure = new ArrayList<>();
			for(int i = 0; i < saveNV; ++i) structure.add(new ArrayList<>());
			for(int i = 0; i < tstructure.size(); ++i){
				for(int pa = 0; pa < tstructure.get(i).size(); ++pa){
					structure.get(MB.get(i)).add(MB.get(tstructure.get(i).get(pa)));
				}
			}
			NV = saveNV;
			NC = saveNC;
			writeStructure(path_to_directory_for_output, j);
			System.out.println("; Complete.");
		}
	}

	public void Classification(ArrayList<int[]> ld, ArrayList<int[]> td, double alpha, ArrayList<ArrayList<Integer>> parentList){
		//init Prob
		prob = new ArrayList<double[]>();
		for(int t = 0; t < td.size(); ++t){
			double[] c = new double[NC[NV - 1]];
			for(int v = 0; v < NC[NV - 1]; ++v){
				c[v] = 0.0;
			}
			prob.add(c);
		}

		ArrayList<HashMap<BigInteger, int[]>> n = makeCft(parentList, ld);
		for(int t = 0; t < td.size(); ++t){
			int[] currentTestData = new int[NV];
			for(int i = 0; i < NV; ++i) currentTestData[i] = td.get(t)[i];
			double[] p = new double[NC[NV - 1]];
			for(int i = 0; i < NC[NV - 1]; ++i) p[i] = 0.0;

			int i = 0;
			for(i = 0; i < NV; ++i){
				double alpha_ij = alpha;
				double denominator = 1.0;
				for(int j = 0; j < parentList.get(i).size(); ++j){
					denominator *= NC[parentList.get(i).get(j)];
				}
				alpha_ij /= denominator;
				double alpha_ijk = alpha / (denominator*NC[i]);

				for(int v = 0; v < NC[NV - 1]; ++v){
					currentTestData[NV - 1] = v;
					//System.out.println("makekey");
					BigInteger hash = makeKey(currentTestData, i, parentList);
					//System.out.println(hash);
					double n_ij = 0.0;
					for(int k = 0; k < NC[i]; ++k) n_ij += (n.get(i).containsKey(hash)) ? n.get(i).get(hash)[k] : 0;
					//if(v == 0) System.out.println(n_ij);
					double n_ijk = (n.get(i).containsKey(hash)) ? n.get(i).get(hash)[currentTestData[i]] : 0;
					p[v] += Math.log(n_ijk + alpha_ijk) - Math.log(n_ij + alpha_ij);
				}
			}
			double total = 0.0;
			for(int c = 0; c < NC[NV - 1]; ++c){
				total += Math.exp(p[c]);
				if(total == 0){
					total = 1;
				}
			}
			for(int j = 0; j < NC[NV - 1]; ++j){
				prob.get(t)[j] = Math.exp(p[j]) / total;
			}
		}
	}

	private double calcCorrect(ArrayList<int[]> td){
		double count = 0;
		for(int t = 0; t < td.size(); ++t){
			double tPred = 0.0;
			int testlabel = td.get(t)[NV - 1];
			boolean flag = false;
			for(int v = 0; v < testlabel; ++v){
				if(prob.get(t)[v] > prob.get(t)[testlabel]){
					flag = true;
				}else if(prob.get(t)[v] == prob.get(t)[testlabel]){
					++tPred;
				}
			}
			for(int v = testlabel + 1; v < NC[NV - 1]; ++v){
				if(prob.get(t)[v] > prob.get(t)[testlabel]){
					flag = true;
				}else if(prob.get(t)[v] == prob.get(t)[testlabel]){
					++tPred;
				}
			}
			if(!flag){
				if(tPred == 0){
					++count;
				}else{
					count += 1.0 / (tPred + 1.0);
				}
			}
		}
		return count;
	}

	private ArrayList<HashMap<BigInteger, int[]>> makeCft(ArrayList<ArrayList<Integer>> parents, ArrayList<int[]> ld){
		ArrayList<HashMap<BigInteger, int[]>> n = new ArrayList<HashMap<BigInteger, int[]>>();
		for(int i = 0; i < NV; ++i) n.add(new HashMap<BigInteger, int[]>());
		for(int l=0; l<ld.size(); l++){
			int[] currentLearnData = new int[NV];
			for(int i = 0; i < NV; ++i) currentLearnData[i] = ld.get(l)[i];

			for(int i = 0; i < NV; i++){
				BigInteger hash = makeKey(currentLearnData, i, parents);
				if(n.get(i).containsKey(hash)){
					(n.get(i).get(hash)[currentLearnData[i]])++;
				}else{
					int[] t = new int[NC[i]];
					for(int j = 0; j < NC[i]; ++j){
						t[j] = 0;
					}
					t[currentLearnData[i]] = 1;
					n.get(i).put(hash, t);
				}
			}
		}
		return n;
	}

	private BigInteger makeKey(int[] data, int i, ArrayList<ArrayList<Integer>> parentList){
		if(!parentList.get(i).isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parentList.get(i).size(); ++p){
				int t = parentList.get(i).get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	private BigInteger product(int a, int b){
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}

	protected void readData(String DataBaseName) throws IOException{
		BufferedReader br = new BufferedReader(new FileReader(new File(DataBaseName)));
		ArrayList<String[]> tRD = new ArrayList<String[]>();
		String line = br.readLine();
		NV = line.split(",").length;
		while(line != null){
			String[] data = line.split(",").clone();
			tRD.add(data);
			line = br.readLine();
		}
		int[] max = new int[NV];
		RD = new ArrayList<int[]>();
		for(int i = 0; i < tRD.size(); ++i){
			RD.add(new int[NV]);
			for(int j = 0; j < NV; ++j){
				RD.get(i)[j] = Integer.parseInt(tRD.get(i)[j]);
				if(max[j] < RD.get(i)[j]){
					max[j] = RD.get(i)[j];
				}
			}
		}
		MNC = 0;
		NC = new int[NV];
		for(int i = 0; i < NV; ++i){
			NC[i] = max[i] + 1;
			if(MNC < NC[i]){
				MNC = NC[i];
			}
		}
		br.close();
	}

	protected void readLearningData(int j) throws IOException{
		if(CrossValidation == 1){
			LD = RD;
			return;
		}
		LD = new ArrayList<int[]>();
		if(j < CrossValidation - 1){
			for(int i = 0; i < RD.size(); ++i){
				if(i < j*(RD.size()/CrossValidation) || (j+1)*(RD.size()/CrossValidation) <= i)
					LD.add(RD.get(i));
			}
		}
		else{
			for(int i = 0; i < (CrossValidation - 1)*(RD.size()/CrossValidation); ++i){
				LD.add(RD.get(i));
			}
		}
	}

	public void writeStructure(String path, int j){
		try{
			File outputDir = new File(path);
			if (outputDir.exists()){
				File file = new File(path + "/" + (j+1) + ".txt");
				file.createNewFile();
				PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(file)));
				for(int i = 0; i < NV; ++i){
					pw.print(i);
					pw.print("\t");
				}
				pw.println();
				for(int i = 0; i < NV; ++i){
					ArrayList<Integer> t = structure.get(i);
					if(t.size() != 0){
						pw.print(t.get(0));
						for(int p = 1; p < t.size(); ++p){
							pw.print(",");
							pw.print(t.get(p));
						}
					}else{
						pw.print(-1);
					}
					pw.print("\t");
				}
				pw.close();
			}else{
				System.out.println();
				System.out.println("The output directory not found.");
				System.exit(0);
			}
		}catch(IOException e){
			System.out.println(e);
		}
	}
}
