import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;

import org.apache.commons.math3.special.Gamma;

public class getLocalScore{
	private int fANB; //learning with ANB constraint if fABN = 1, learning GBN otherwise.
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public ArrayList<int[]> LD; //Learning data.
	public double alpha; // equivalent sample size.
	public ArrayList<Double[]> LS; // Local Score List.

	public getLocalScore(int fANB, ArrayList<int[]> ld, int nv, int[] nc, int mnc, double alpha){
		this.fANB = fANB;
		LD = ld;
		NV = nv;
		NC = nc;
		MNC = mnc;
		this.alpha = alpha;
		initLS();
		setLocalScores();
	}

	private void initLS(){
		LS = new ArrayList<>();
		for(int i = 0; i < NV - fANB; ++i){
			LS.add(new Double[(int) Math.pow(2, NV - fANB - 1)]);
		}
	}

	private void setLocalScores(){
		for(int i = 0; i < NV - fANB; ++i){
			ArrayList<Integer> parents = new ArrayList();
			for(int c = 0; c < i; ++c) parents.add(c);
			for(int c = i; c < NV - fANB - 1; ++c) parents.add(c+1);
			if(fANB == 1) parents.add(NV - 1);
			Map<BigInteger, int[]> cft = new HashMap<>();
			setCft(cft, parents, i);
			double denominator = 1.0;
			for(int c = 0; c < parents.size(); ++c){
				denominator *= NC[parents.get(c)];
			}
			int j = (1 << (NV - fANB - 1)) - 1;
			LS.get(i)[j] = calcLS(cft, i, denominator);
			ForkJoinPool pool = new ForkJoinPool();
			pool.invoke(new LocalScoreDP(cft, i, 0, j, denominator));
		}
	}


	private class LocalScoreDP extends RecursiveAction {
		private Map<BigInteger, int[]> cft;
		private int child;
		private int p;
		private int j;
		private double denominator;
		LocalScoreDP(Map<BigInteger, int[]> cft, int child, int p, int j, double denominator) {
			this.cft = cft;
			this.child = child;
			this.p = p;
			this.j = j;
			this.denominator = denominator;
		}
		@Override
		protected void compute() {
			if(p >= NV - fANB - 1){
				return;
			}
			LocalScoreDP f1 = new LocalScoreDP(cft, child, p + 1, j, denominator);
			f1.fork();
			int margi = (p >= child) ? p + 1 : p;
			Map<BigInteger, int[]> cft2 = new HashMap();
			for(Map.Entry<BigInteger, int[]> e : cft.entrySet()) {
				BigInteger key = e.getKey();
				BigInteger prod1 = product(MNC, p);
				BigInteger prod2 = prod1.multiply(BigInteger.valueOf(MNC));
				BigInteger down = key.remainder(prod1);
				BigInteger up = key.divide(prod2);
				key = up.multiply(prod2).add(down);

				if(cft2.containsKey(key)){
					for(int index = 0; index < NC[child]; ++index){
						cft2.get(key)[index] += e.getValue()[index];
					}
				}else{
					int[] t = e.getValue().clone();
					cft2.put(key, t);
				}
			}

			int j2 = j - (1 << p);
			denominator /= NC[margi];
			LS.get(child)[j2] = calcLS(cft2, child, denominator);
			LocalScoreDP f2 = new LocalScoreDP(cft2, child, p + 1, j2, denominator);
			f2.invoke();
			f1.join();
		}
	}

	private void setCft(Map<BigInteger, int[]> cft, ArrayList<Integer> parents, int i){
		for(int l = 0; l < LD.size(); ++l){
			BigInteger key = makeKey(LD.get(l), parents);
			int index = LD.get(l)[i];
			if(cft.containsKey(key)){
				cft.get(key)[index]++;
			}else{
				int[] t = new int[NC[i]];
				for(int j = 0; j < NC[i]; ++j){
					t[j] = 0;
				}
				t[index] = 1;
				cft.put(key, t);
			}
		}
	}

	private double calcLS(Map<BigInteger, int[]> cft, int i, double denominator){
		double L = 0.0;
		double alpha_ij = alpha;
		alpha_ij /= denominator;
		for(Map.Entry<BigInteger, int[]> e : cft.entrySet()) {
			double n_ij = 0;
			for(int k = 0; k < NC[i]; ++k){
				double alpha_ijk = alpha_ij / (double)NC[i];
				double n_ijk = e.getValue()[k];
				L += Gamma.logGamma(alpha_ijk + n_ijk) - Gamma.logGamma(alpha_ijk);
				n_ij += n_ijk;
			}
			L += Gamma.logGamma(alpha_ij) - Gamma.logGamma(alpha_ij + n_ij);
		}
		return L;
	}

	private BigInteger makeKey(int[] data, ArrayList<Integer> parents){ // 親状態 -> 数値 に変換する
		if(!parents.isEmpty()){
			BigInteger key = BigInteger.ZERO;
			for(int p = 0; p < parents.size(); ++p){
				int t = parents.get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	public BigInteger product(int a, int b){// a^b
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}
}