import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveAction;

import org.apache.commons.math3.distribution.GammaDistribution;

public class getLocalScore_aCLL{
	private int fANB; //learning with ANB constraint if fABN = 1, learning GBN otherwise.
	public int NV; //Number of Variables.
	public int[] NC; //NC[i] : the i-th variable configures.
	public int MNC; // max{NC[i]}
	public ArrayList<int[]> LD; //Learning data.
	public ArrayList<Double[]> LS; // Local Score List.
	public double BETA;

	public getLocalScore_aCLL(int fANB, ArrayList<int[]> ld, int nv, int[] nc, int mnc, double alpha){
		this.fANB = fANB;
		LD = ld;
		NV = nv;
		NC = nc;
		MNC = mnc;
		BETA = abMCMC(NC[NV - 1], LD.size());
		initLS();
		setLocalScores_aCLL();
	}

	private void initLS(){
		LS = new ArrayList<>();
		for(int i = 0; i < NV - fANB; ++i){
			LS.add(new Double[(int) Math.pow(2, NV - fANB - 1)]);
		}
	}

	private void setLocalScores_aCLL(){
		for(int i = 0; i < NV - fANB; ++i){
			ArrayList<Integer> parents = new ArrayList();
			for(int c = 0; c < i; ++c) parents.add(c);
			for(int c = i; c < NV - fANB - 1; ++c) parents.add(c+1);
			if(fANB == 1) parents.add(NV - 1);
			Map<BigInteger, int[][]> cft = new HashMap<>();
			setCftaCLL(cft, parents, i);
			double denominator = 1.0;
			for(int c = 0; c < parents.size(); ++c){
				denominator *= NC[parents.get(c)];
			}
			int j = (1 << (NV - fANB - 1)) - 1;
			LS.get(i)[j] = calcLSaCLL(cft, i, denominator);


			//recursiveLS(cft, i, 0, j, denominator);


			ForkJoinPool pool = new ForkJoinPool();
			pool.invoke(new LocalScoreDPaCLL(cft, i, 0, j, denominator));
		}
	}


	private class LocalScoreDPaCLL extends RecursiveAction {
		private Map<BigInteger, int[][]> cft;
		private int child;
		private int p;
		private int j;
		private double denominator;

		LocalScoreDPaCLL(Map<BigInteger, int[][]> cft, int child, int p, int j, double denominator) {
			this.cft = cft;
			this.child = child;
			this.p = p;
			this.j = j;
			this.denominator = denominator;
		}

		@Override
		protected void compute() {
			if(p >= NV - fANB - 1){
				return;
			}
			LocalScoreDPaCLL f1 = new LocalScoreDPaCLL(cft, child, p + 1, j, denominator);
			f1.fork();
			//recursiveLS(cft, child, p + 1, j, denominator);
			int margi = (p >= child) ? p + 1 : p;
			Map<BigInteger, int[][]> cft2 = new HashMap();


			for(Map.Entry<BigInteger, int[][]> e : cft.entrySet()) {
				BigInteger key = e.getKey();
				BigInteger prod1 = product(MNC, p);
				BigInteger prod2 = prod1.multiply(BigInteger.valueOf(MNC));
				BigInteger down = key.remainder(prod1);
				BigInteger up = key.divide(prod2);
				key = up.multiply(prod2).add(down);

				//int index = (int)(e.getKey() - up*MNC+down) / (int)Math.pow(MNC, posi);
				if(cft2.containsKey(key)){
					for(int c = 0; c < NC[NV - 1]; ++c){
						for(int index = 0; index < NC[child]; ++index){
							cft2.get(key)[c][index] += e.getValue()[c][index];
						}
					}
				}else{
					int[][] t = new int[NC[NV - 1]][NC[child]];
					for(int c = 0; c < NC[NV - 1]; ++c){
						t[c] = e.getValue()[c].clone();
					}
					cft2.put(key, t);
				}
			}


			int j2 = j - (1 << p);
			if(isSmallerNumberParent(j2, 3)){
				LS.get(child)[j2] = calcLSaCLL(cft2, child, denominator);
			}else{
				LS.get(child)[j2] = 0.0;
			}
			LocalScoreDPaCLL f2 = new LocalScoreDPaCLL(cft2, child, p + 1, j2, denominator);
			f2.invoke();
			f1.join();
			//recursiveLS(cft2, child, p + 1, j2, denominator);
		}
	}

	private boolean isSmallerNumberParent(int j, int num){ // return (The number of parents < num) ? 1 : 0
		int r = fANB;
		for(int i = 0; i < NV - fANB - 1; ++i){
			r += (j >> i) & 1;
		}
		return (r < num);
	}

	private void setCftaCLL(Map<BigInteger, int[][]> cft, ArrayList<Integer> parents, int i){
		for(int l = 0; l < LD.size(); ++l){
			BigInteger key = makeKeyaCLL(LD.get(l), parents, i);
			int index = LD.get(l)[i];
			int c = LD.get(l)[NV - 1];
			if(cft.containsKey(key)){
				cft.get(key)[c][index]++;
			}else{
				int[][] t = new int[NC[NV - 1]][NC[i]];
				for(int j1 = 0; j1 < NC[NV - 1]; ++j1){
					for(int j2 = 0; j2 < NC[i]; ++j2){
						t[j1][j2] = 0;
					}
				}
				t[c][index] = 1;
				cft.put(key, t);
			}
		}
	}

	private double calcLSaCLL(Map<BigInteger, int[][]> cft, int i, double denominator){
		double alpha_i = 1.0, L = 0.0;
		double alpha_ij = alpha_i;
		alpha_ij /= denominator;

		double pseudo = 5.0;
		double beta = BETA;
		//double beta = coefficient[NC[NV - 1]];
		double alpha = 1 + beta;
		double penalty = 0.0;
		for(Map.Entry<BigInteger, int[][]> e : cft.entrySet()) {
			double[][] bunsi_ijck = new double[NC[NV - 1]][NC[i]];

			for(int c = 0; c < NC[NV - 1]; ++c){
				for(int k = 0; k < NC[i]; ++k){
					double n_ijk = 0;
					for(int c1 = 0; c1 < NC[NV - 1]; ++c1){
						n_ijk += e.getValue()[c1][k];
					}
					double n_ijck = e.getValue()[c][k];
					bunsi_ijck[c][k] = alpha*n_ijck + beta*(n_ijk - n_ijck);
				}
			}
			for(int c = 0; c < NC[NV - 1]; ++c){
				double bunbo = 0;
				for(int k = 0; k < NC[i]; ++k){
					double t = bunsi_ijck[c][k];
					if(t < pseudo){
						t = pseudo;
					}
					bunbo += t;
				}
				for(int k = 0; k < NC[i]; ++k){
					double t = bunsi_ijck[c][k];
					if(t < pseudo){
						t = pseudo;
					}
					L += bunsi_ijck[c][k] * (Math.log(t) - Math.log(bunbo));
				}
			}
		}
		return L;
	}

	private BigInteger makeKeyaCLL(int[] data, ArrayList<Integer> parents, int i){ // 親状態 -> 数値 に変換する
		if(!parents.isEmpty()){
			BigInteger key = BigInteger.ZERO;
			int one = (i != NV - 1) ? 1 : 0;
			for(int p = 0; p < parents.size() - one; ++p){
				int t = parents.get(p);
				key = key.add(product(MNC, p).multiply(BigInteger.valueOf(data[t])));
			}
			return key;
		}else{
			return BigInteger.ZERO;
		}
	}

	public BigInteger product(int a, int b){// a^b
		BigInteger r = BigInteger.ONE;
		for(int i = 0; i < b; ++i){
			r = r.multiply(BigInteger.valueOf(a));
		}
		return r;
	}

	private double abMCMC(int s, int dataSize){
		int m = 10000000;
		double a = 1;
		double b = dataSize;

		GammaDistribution aGD = new GammaDistribution(a, 1);
		GammaDistribution bGD = new GammaDistribution(b, 1);
		double[] xpoint = new double[m];
		double[] ypoint = new double[m];
		double xave = 0;
		double yave = 0;
		for(int d = 0; d < m; ++d){
			double sum = 0;
			double[] u = new double[s];
			for(int c = 0; c < s; ++c){
				u[c] = aGD.sample();
				sum += u[c];
			}
			sum += bGD.sample();

			xpoint[d] = 0;
			ypoint[d] = 0;
			for(int c = 0; c < s; ++c){
				double uc = u[c]/sum;
				xpoint[d] += Math.log(uc);
				ypoint[d] += uc;
			}
			ypoint[d] = -Math.log(ypoint[d]);

			xave += xpoint[d];
			yave += ypoint[d];
		}
		xave /= m;
		yave /= m;

		// linear regression
		double denominator = 0;
		double numerator = 0;
		for(int d = 0; d < m; ++d){
			numerator += (xpoint[d] - xave) * (ypoint[d] - yave);
			denominator += (xpoint[d] - xave) * (xpoint[d] - xave);
		}
		double beta = numerator / denominator;
		double gamma = yave - beta * xave;
		return beta;
	}
}